#!/bin/sh

# Installer utility functions



# Simple logging routine; echoes arguments to "globally" defined LogFileFullPath,
# or stdout if no log file set up, or if log file does not exist.

function Log
	{
	declare -r TimeStamp=`date "+%H:%M"`
	
	if [ -n "${gLogFileFullPath}" ] ; then
		${ECHO} "[$$ ${TimeStamp}]  " $@ >> "${gLogFileFullPath}"
	else
		${ECHO} "[$$ ${TimeStamp}]  " $@
	fi
	}



# Examine shell error result; report error and continue if non-zero.
# Includes any parameters as part of error message.  Returns saved shell
# error result as function result.

function WarnIfError
{
	local Err=$?
	if [ "${Err}" -ne 0 ]; then
		Log "Warning: error $Err: $*"
	fi
	return "${Err}"
}



# Examine shell error result; report error and exit if non-zero.
# Includes any parameters as part of error message.

function FailIfError
{
	local Err=$?
	if [ "${Err}" -ne 0 ]; then
		Log "Error $Err: $*"
		exit ${Err}
	fi
}



# Report an error and exit.  First parameter must be error code.
# Includes any additional parameters as part of error message.

function FailWithError
{
	declare -i Err="$1";
	declare Msg="";
	
    if [ $# -lt 1 ] ; then
        Msg="FailWithError: missing argument"
        Err=1
    else
    	Err="$1"
    	shift
    	Msg="$*"
    fi
    
    Log "$Msg"
	exit ${Err}
}



# Compares actual argument count with expected, and fails if not equal, writing error
# message to log, and exiting with error 1.
# 
# Parameters:
# 
#    1: name of routine whose argument list is being checked
#    2: number of arguments passed to routine (normally called with actual parameter of #$)
#    3: correct number of arguments:
#       if negative, argument count must be >= absolute value 
#       if positive, argument count must = value

function CheckArgumentCount
	{
	declare -r FunctionName="$1"
	declare -r ActualArgCount="$2"
	declare -r DesiredArgCount="$3"
	
	if [ $# -eq 3 ] ; then
		if [ "${DesiredArgCount}" -lt 0 ] ; then
			declare -r MinArgCount=-${DesiredArgCount}
			if [ "${MinArgCount}" -gt "${ActualArgCount}" ] ; then
				FailWithError 1 "${FunctionName}: wrong number of arguments ${ActualArgCount}, should be at least ${MinArgCount}"
			fi
		else
			if [ "${DesiredArgCount}" -ne "${ActualArgCount}" ] ; then
				FailWithError 1 "${FunctionName}: wrong number of arguments ${ActualArgCount}, should be ${DesiredArgCount}"
			fi
		fi
	else
		FailWithError 1 "CheckArgumentCount: called with wrong number of arguments(!)"
	fi
	}



# Copies files from source to destination.  Continues regardless of
# errors.  Copies resource forks.

function CopyIgnoringErrors
{
    declare -r SOURCE="$1"
    declare -r DEST="$2"
    
    CheckArgumentCount "CopyIgnoringErrors" $# 2

    Log "Copying ${SOURCE} to ${DEST}..."
    if [ -e "${SOURCE}" ] ; then
        Log "${DITTO} -rsrcFork \"${SOURCE}\" \"${DEST}\" "
        ${DITTO} "${SOURCE}" "${DEST}" 
    else    
        Log "${SOURCE} not found."
    fi
}



# Copies files from source to destination.  Script exits if files
# not present or error in copying.  Copies resource forks.

function CopyOrFail
{
    declare -r SOURCE="$1"
    declare -r DEST="$2"

    CheckArgumentCount "CopyOrFail" $# 2
    
    ${DITTO} -rsrcFork "${SOURCE}" "${DEST}"
    FailIfError "copying files from ${SOURCE} to ${DEST}"
}



# Accepts two parameters: first is full path to item for which the numerical, octal
# chmod-compatible mode is desired.  Second is the *name* of the variable in caller
# scope in which to return the result.
# 
# The result is a *string* that looks like an octal value, exactly equivalent to
# what chmod would require in order to set permissions to their current state.
# 
# The result value will be placed in a caller variable having the specified name.
# unfortunately, this name cannot match the name of any local variable used in this
# routine.
# 
# *** IMPORTANT: this return-value mechanism is not documented or condoned in any
#     known literature, and may not be entirely reliable; needs further testing /
#     verification / perusal of bash source.
# 
# This routine has been exhaustively tested on all possible permission settings and
# is known to be 100% correct. 

function GetFSItemMode
	{
	declare -i _ItemModeDec=0
	declare _FileNameAndPath="$1"
	declare _ResultVarName="$2"
	
    CheckArgumentCount "GetFSItemMode" $# 2
    
	declare _LSOutput=`ls -al "${_FileNameAndPath}"`
#	Log "_LSOutput: ${_LSOutput}"
	
	declare _PermsField=${_LSOutput:0:10}
	
	declare _OwnerPerms=${_PermsField:1:3}
	declare _GroupPerms=${_PermsField:4:3}
	declare _OtherPerms=${_PermsField:7:3}
	
	declare _OwnerR=${_OwnerPerms:0:1}
	declare _OwnerW=${_OwnerPerms:1:1}
	declare _OwnerX=${_OwnerPerms:2:1}
	
	declare _GroupR=${_GroupPerms:0:1}
	declare _GroupW=${_GroupPerms:1:1}
	declare _GroupX=${_GroupPerms:2:1}
	
	declare _OtherR=${_OtherPerms:0:1}
	declare _OtherW=${_OtherPerms:1:1}
	declare _OtherX=${_OtherPerms:2:1}
	
	declare _Dummy
	
	if [ "${_OwnerX}" = "x" ] ; then
		_ItemModeDec=$_ItemModeDec+64
	elif [ "${_OwnerX}" = "-" ] ; then
		:
	elif [ "${_OwnerX}" = "s" ] ; then
		_ItemModeDec=$_ItemModeDec+2112
	elif [ "${_OwnerX}" = "S" ] ; then
		_ItemModeDec=$_ItemModeDec+2048
	fi
	if [ "${_OwnerW}" = "w" ] ; then
		_ItemModeDec=$_ItemModeDec+128
	fi
	if [ "${_OwnerR}" = "r" ] ; then
		_ItemModeDec=$_ItemModeDec+256
	fi

	if [ "${_GroupX}" = "x" ] ; then
		_ItemModeDec=$_ItemModeDec+8
	elif [ "${_GroupX}" = "-" ] ; then
		:
	elif [ "${_GroupX}" = "s" ] ; then
		_ItemModeDec=$_ItemModeDec+1032
	elif [ "${_GroupX}" = "S" ] ; then
		_ItemModeDec=$_ItemModeDec+1024
	fi
	if [ "${_GroupW}" = "w" ] ; then
		_ItemModeDec=$_ItemModeDec+16
	fi
	if [ "${_GroupR}" = "r" ] ; then
		_ItemModeDec=$_ItemModeDec+32
	fi

	if [ "${_OtherX}" = "x" ] ; then
		_ItemModeDec=$_ItemModeDec+1
	elif [ "${_OtherX}" = "-" ] ; then
		:
	elif [ "${_OtherX}" = "t" ] ; then
		_ItemModeDec=$_ItemModeDec+513
	elif [ "${_OtherX}" = "T" ] ; then
		_ItemModeDec=$_ItemModeDec+512
	fi
	if [ "${_OtherW}" = "w" ] ; then
		_ItemModeDec=$_ItemModeDec+2
	fi
	if [ "${_OtherR}" = "r" ] ; then
		_ItemModeDec=$_ItemModeDec+4
	fi

#	Log "_ItemModeDec: ${_ItemModeDec}"
	
	declare -i _OtherModeOctal=$(( _ItemModeDec % 8 ))
	declare -i _GroupModeOctal=$(( (_ItemModeDec >> 3) % 8 ))
	declare -i _OwnerModeOctal=$(( (_ItemModeDec >> 6) % 8 ))
	declare -i _SpecialModeOctal=$(( (_ItemModeDec >> 9) % 8 ))
	
	declare _OctalItemMode=`echo -n "${_SpecialModeOctal}${_OwnerModeOctal}${_GroupModeOctal}${_OtherModeOctal}"`
#	Log "_OctalItemMode: ${_OctalItemMode}"
	
	export "${_ResultVarName}"="${_OctalItemMode}"
	}



####
# getOwner inPath
# Gets the owner of the specified file or folder
#   inPath  the file or folder to get the owner of
#   RESULT  is owner if it can be retrieved and unset otherwise
#   return  0 on success, non-zero otherwise
function getOwner
{
	local RETURN=0
	RESULT=
	
	if [ $# -ne 1 ] ; then
		Log "getOwner: wrong number of arguments."
		return 2
	fi
	
	local IN_PATH="$1"
	
	if [ -e "${IN_PATH}" ] ; then
		RESULT=`${LS} -ldn "${IN_PATH}" | ${AWK} '{ ORS=""; print $3; }' -`
		RETURN=$?
		if [ ${RETURN} -ne 0 ] ; then
			Log "getOwner: Unable to get owner for ${IN_PATH}."
			RESULT=
		else
			if [ -z "${RESULT}" ] ; then
				Log "getOwner: Owner is empty for ${IN_PATH}."
				RESULT=
				RETURN=3
			fi
		fi
	else
		Log "getOwner: Unable to access ${IN_PATH} to get owner"
		RETURN=4
	fi
	
	return ${RETURN}
}



####
# getGroup inPath
# Gets the group membership of the specified file or folder
#   inPath  the file or folder to get the group of
#   RESULT  group if it can be retrieved and unset otherwise
#   return  0 on success, non-zero otherwise
function getGroup
{
	local RETURN=0
	RESULT=
	
	if [ $# -ne 1 ] ; then
		Log "getGroup: wrong number of arguments."
		return 2
	fi
	
	local IN_PATH="$1"
	
	if [ -e "${IN_PATH}" ] ; then
		RESULT=`${LS} -ldn "${IN_PATH}" | ${AWK} '{ ORS=""; print $4; }' -`
		RETURN=$?
		if [ ${RETURN} -ne 0 ] ; then
			Log "getGroup: Unable to get group for ${IN_PATH}."
			RESULT=
		else
			if [ -z "${RESULT}" ] ; then
				Log "getGroup: Group is empty for ${IN_PATH}."
				RESULT=
				RETURN=3
			fi
		fi
	else
		Log "getGroup: Unable to access ${IN_PATH} to get group"
		RETURN=4
	fi
	
	return ${RETURN}
}



####
# getOwnerGroup inPath
# Get the owner and group of the specified file or folder in "owner:group"
# form suitable for use in calls to 'chown'
#   inPath  the file or folder to get the owner and group of
#   RESULT  "owner:group" if it can be retrieved and unset otherwise
#   return  0 on success, non-zero otherwise
function getOwnerGroup
{
	local RETURN=0
	RESULT=
	
	if [ $# -ne 1 ] ; then
		Log "getOwnerGroup: wrong number of arguments."
		return 2
	fi
	
	local IN_PATH="$1"
	
	if [ -e "${IN_PATH}" ] ; then
		RESULT=`${LS} -ldn "${IN_PATH}" | ${AWK} '{ ORS=""; print $3":"$4; }' -`
		RETURN=$?
		if [ ${RETURN} -ne 0 ] ; then
			Log "getOwnerGroup: Unable to get owner & group for ${IN_PATH}."
			RESULT=
		else
			if [ -z "${RESULT}" -o "${RESULT}" = ":" ] ; then
				Log "getOwnerGroup: Owner & group is empty for ${IN_PATH}."
				RESULT=
				RETURN=3
			fi
		fi
	else
		Log "getOwnerGroup: Unable to access ${IN_PATH} to get owner and group"
		RETURN=4
	fi
	
	return ${RETURN}
}



####
# setInstalledOwnership inBOMPath inInstallFolder inOwnerGroup
# 
# Sets the owner and group membership of files installed with the specified BOM
# to match inOwnerGroup.  The files must have already been installed to
# inInstallFolder.
#   inBOMPath       path to Archive.BOM that was already installed
#   inInstallFolder path to folder where archive was installed
#   inOwnerGroup    ownership & group to apply to installed files (i.e. root:admin)
#   RESULT          unset
#   return          0 on success, non-zero otherwise
function setInstalledOwnership
{
	local RETURN=0
	RESULT=
	
	if [ $# -ne 3 ] ; then
		Log "setInstalledOwnership: wrong number of arguments."
		return 2
	fi
	
	local IN_BOM_PATH="$1"
	local IN_INSTALL_FOLDER="$2"
	local IN_OWNER_GROUP="$3"
	
	if [ ! -r "${IN_BOM_PATH}" ] ; then
		Log "setInstalledOwnership: cannot read BOM at ${IN_BOM_PATH}."
		return 1
	fi
	
	if [ ! -r "${IN_INSTALL_FOLDER}" ] ; then
		Log "setInstalledOwnership: cannot access install path at ${IN_INSTALL_FOLDER}."
		return 3
	fi
	
	if [ -z "${IN_OWNER_GROUP}" ] ; then
		Log "setInstalledOwnership: owner and group cannot be empty."
		return 4
	fi
	
	# change current directory to install dir
	pushd "${IN_INSTALL_FOLDER}" >> /dev/null
	RETURN=$?
	if [ ${RETURN} -ne 0 ] ; then
		Log "setInstalledOwnership: pushd returned an error."
		return ${RETURN}
	fi
	
	# get installed file paths
	# skip resource fork names since all will have associated data forks
	# replace newlines with \0
	# run xargs -0 to pass files to chown
	${LSBOM} -s "${IN_BOM_PATH}" \
		| ${GREP} --invert-match '/\._' \
		| ${AWK} ' { ORS=""; printf "%s%c", $0, 0; } ' - \
		| ${XARGS} -0 ${CHOWN} "${IN_OWNER_GROUP}"
	RETURN=$?
	if [ ${RETURN} -ne 0 ] ; then
		Log "setInstalledOwnership: an error occurred setting ownership (${RETURN})."
	fi
	
	# restore prior current directory
	popd >> /dev/null
	RETURN=$?
	if [ ${RETURN} -ne 0 ] ; then
		Log "setInstalledOwnership: popd returned an error."
		return ${RETURN}
	fi
	
	return ${RETURN}
}



####
# setDittoOwnership inDittoSource inDittoDest inOwnerGroup
# 
# Sets the owner and group membership of files previously copied with ditto
# to match inOwnerGroup.  Note: if the contents of inDittoSource have changed
# since the ditto command was run, the results are undefined.
#   inDittoSource   source used in previous ditto command
#   inDittoDest     destination used in previous ditto command
#   inOwnerGroup    ownership & group to apply to copied files (i.e. root:admin)
#   RESULT          unset
#   return          0 on success, non-zero otherwise
function setDittoOwnership
{
	local RETURN=0
	RESULT=
	
	if [ $# -ne 3 ] ; then
		Log "setDittoOwnership: wrong number of arguments."
		return 2
	fi
	
	local IN_DITTO_SOURCE="$1"
	local IN_DITTO_DEST="$2"
	local IN_OWNER_GROUP="$3"
	
	if [ ! -r "${IN_DITTO_SOURCE}" ] ; then
		Log "setDittoOwnership: cannot read source at ${IN_DITTO_SOURCE}."
		return 1
	fi
	
	if [ ! -r "${IN_DITTO_DEST}" ] ; then
		Log "setDittoOwnership: cannot access destination path at ${IN_DITTO_DEST}."
		return 3
	fi
	
	if [ -z "${IN_OWNER_GROUP}" ] ; then
		Log "setDittoOwnership: owner and group cannot be empty."
		return 4
	fi
	
	# if the ditto destination is a relative path
	# the stick the current working directory on the front
	if [ -z "`${ECHO} \"${IN_DITTO_DEST}\" | ${EGREP} \"^/\"`" ] ; then
		Log "Destination path is relative, prepending current directory."
		local CURRENT_DIR=`pwd -P`
		RETURN=$?
		if [ ${RETURN} -eq 0 ] ; then
			IN_DITTO_DEST="${CURRENT_DIR}/${IN_DITTO_DEST}"
			Log "New destination path is ${IN_DITTO_DEST}."
		else
			Log "setDittoOwnership: error getting pwd."
			return ${RETURN}
		fi
	fi
	
	if [ -d "${IN_DITTO_SOURCE}" ] ; then
		# it's a directory, set destination owner:group for each source item
		
		# change current directory to ditto source dir
		pushd "${IN_DITTO_SOURCE}" >> /dev/null
		RETURN=$?
		if [ ${RETURN} -ne 0 ] ; then
			Log "setDittoOwnership: pushd returned an error."
			return ${RETURN}
		fi
		
		# find source files
		# set ownership on same paths in ditto dest
		${FIND} . -print0 \
			| (cd "${IN_DITTO_DEST}"; \
			${XARGS} -0 ${CHOWN} "${IN_OWNER_GROUP}")
		RETURN=$?
		if [ ${RETURN} -ne 0 ] ; then
			Log "setDittoOwnership: an error occurred setting ownership (${RETURN})."
		fi
		
		# restore prior current directory
		popd >> /dev/null
		RETURN=$?
		if [ ${RETURN} -ne 0 ] ; then
			Log "setDittoOwnership: popd returned an error."
		fi
	else
		# it's just a file, set destination file owner:group
		${CHOWN} "${IN_OWNER_GROUP}" "${IN_DITTO_DEST}"
	fi
	
	return ${RETURN}
}



####
# installOEM inOEMSource inOEMDest
# Copies the contents of inOEMSource to the folder inOEMDest
# propagating ownership and group from the destination folder
#   inOEMSource OEM_ folder to copy from
#   inOEMDest   folder where OEM_ content should be installed
#   RESULT      unset
#   return      0 on success, non-zero otherwise
function installOEM
{
	local THE_OWNER_GROUP=""
	local RETURN=0
	RESULT=
	
	if [ $# -ne 2 ] ; then
		Log "installOEM: wrong number of arguments."
		return 2
	fi
	
	local IN_OEM_Source="$1"
	local IN_OEM_Dest="$2"
	local OEM_Temp="/tmp/OEM_"
	
	# validate IN_OEM_Source
	if [ ! -d "${IN_OEM_Source}" ] ; then
		Log "installOEM: source folder ${IN_OEM_Source} does not exist."
		return 3
	fi
	
	# validate IN_OEM_Dest
	if [ ! -d "${IN_OEM_Dest}" ] ; then
		Log "installOEM: destination folder ${IN_OEM_Dest} does not exist."
		return 4
	fi
	
	Log "Getting owner and group for install folder ${IN_OEM_Dest}..."
	getOwnerGroup "${IN_OEM_Dest}"
	RETURN=$?
	if [ ${RETURN} -eq 0 -a -n "${RESULT}" ] ; then
		local THE_OWNER_GROUP="${RESULT}"
	else
		Log "installOEM: Unable to get owner & group for ${IN_OEM_Dest}, skipping..."
	fi
	
	if [ -n "${THE_OWNER_GROUP}" ] ; then
		if [ ! -e "${OEM_Temp}" ] ; then
			Log "Copying OEM content to ${OEM_Temp} to fix up ownership & permissions..."
			${DITTO} -rsrcFork "${IN_OEM_Source}" "${OEM_Temp}"
			Log "Setting owner:group to ${THE_OWNER_GROUP}"
			${CHOWN} -R "${THE_OWNER_GROUP}" "${OEM_Temp}"
			${FIND} "${OEM_Temp}" -print0 | ${XARGS} -0 ${CHMOD} ug+rw
			${FIND} "${OEM_Temp}" -print0 | ${XARGS} -0 ${CHMOD} o+r-w
			${FIND} "${OEM_Temp}" -type d -print0 | ${XARGS} -0 ${CHMOD} ugo+x
			
			Log "Installing OEM content into ${IN_OEM_Dest}"
			${DITTO} -rsrcFork "${OEM_Temp}" "${IN_OEM_Dest}"
			
			if [ "${OEM_Temp}" != "/" ] ; then
				Log "Removing temporary content"
				${RM} -rf "${OEM_Temp}"
			else
				Log "Temp content dir is /, not deleting."
			fi
		else
			Log "OEM fix up directory at ${OEM_Temp} is occupied, skipping..."
		fi
	else
		Log "Owner/group information unavailable, skipping..."
	fi
	
	return ${RETURN}
}



####
# deleteReceipt inTargetVolume inPackage
# Deletes the receipt for the specified package
#   inTargetVolume  path to installation target volume
#   inPackage       name of the package whose receipt should be deleted
#   RESULT          unset
#   return          0 on success, non-zero otherwise
function deleteReceipt
{
	local RETURN=0
	RESULT=
	
	if [ $# -ne 2 ] ; then
		Log "deleteReceipt: wrong number of arguments."
		return 2
	fi
	
	local IN_TARGET_VOLUME="$1"
	local IN_PACKAGE="$2"
	
	if [ ! -e "${IN_TARGET_VOLUME}" ] ; then
		Log "deleteReceipt: target volume does not exist."
		return 2
	fi
	
	if [ -z "${IN_PACKAGE}" ] ; then
		Log "deleteReceipt: package not specified."
		return 2
	fi
	
	local RECEIPT="${IN_TARGET_VOLUME}/Library/Receipts/${IN_PACKAGE}"
	if [ -e "${RECEIPT}" ] ; then
		if [ "${RECEIPT}" != "/" ] ; then
			Log "Removing receipt at ${RECEIPT}..."
			${RM} -rf "${RECEIPT}"
			RETURN=$?
			if [ ${RETURN} -ne 0 ] ; then
				Log "deleteReceipt: Error removing receipt at ${RECEIPT}."
			fi
		else
			# this should never occur because RECEIPT contains /Library/Receipts/
			Log "deleteReceipt: Receipt path is /, skipping deletion."
		fi
	else
		Log "deleteReceipt: Receipt does not exist at ${RECEIPT}, skipping deletion."
	fi
	
	return ${RETURN}
}



####
# dittoWithOwnership inSource inDest inOwnerGroup inParentPermissions
# Uses ditto to copy files from inSource to inDest, setting ownership
# on the copies to inOwnerGroup.  If parent folders need to be created,
# the folders will have the inOwnerGroup and be mode inParentPermissions
#   inSource            ditto source
#   inDest              ditto destination
#   inOwnerGroup        owner:group to set copies to
#   inParentPermissions permission to use if any parent folders need to be created
#   RESULT              unset
#   return              0 on success, non-zero otherwise
function dittoWithOwnership
{
	local RETURN=0
	RESULT=
	
	if [ $# -ne 4 ] ; then
		Log "dittoWithOwnership: wrong number of arguments."
		return 2
	fi
	
	local IN_DITTO_SOURCE="$1"
	local IN_DITTO_DEST="$2"
	local IN_OWNER_GROUP="$3"
	local IN_PARENT_PERMISSIONS="$4"
	
	if [ ! -e "${IN_DITTO_SOURCE}" ] ; then
		Log "dittoWithOwnership: source ${IN_DITTO_SOURCE} does not exist."
		return 2
	fi
	
	if [ -z "${IN_DITTO_DEST}" ] ; then
		Log "dittoWithOwnership: destination not specified."
		return 2
	fi
	
	if [ -z "${IN_OWNER_GROUP}" ] ; then
		Log "dittoWithOwnership: owner & group not specified."
		return 2
	fi
	
	if [ -z "${IN_PARENT_PERMISSIONS}" ] ; then
		Log "dittoWithOwnership: parent permissions not specified."
		return 2
	fi
	
	# precreate path to destination
	createParentWithOwnershipAndPermissions "${IN_DITTO_DEST}" "${IN_OWNER_GROUP}" "${IN_PARENT_PERMISSIONS}"
	RETURN=$?
	if [ ${RETURN} -eq 0 ] ; then
		# ditto the content
		${DITTO} -rsrcFork "${IN_DITTO_SOURCE}" "${IN_DITTO_DEST}"
		RETURN=$?
		if [ ${RETURN} -eq 0 ] ; then
			# set ownership of copied files
			setDittoOwnership "${IN_DITTO_SOURCE}" "${IN_DITTO_DEST}" "${IN_OWNER_GROUP}"
			RETURN=$?
		else
			Log "dittoWithOwnership: error copying content from ${IN_DITTO_SOURCE} to ${IN_DITTO_DEST}."
		fi
	else
		Log "dittoWithOwnership: error creating parent directories."
	fi
	
	return ${RETURN}
}



####
# createParentWithOwnershipAndPermissions inPath inOwnerGroup inPermissions
# Creates the parent path folders with the specified ownership and permissions.
# Ownership and permissions of existing parent folders remains unchanged.  If
# the parent already exists, this does nothing.
#   inPath          path to file/folder whose parent folder path should be created
#   inOwnerGroup    owner:group to use for created parent folders
#   inPermissions   permissions to apply to newly created parent folders
#   RESULT          unset
#   return          0 on success, non-zero otherwise
function createParentWithOwnershipAndPermissions
{
	local RETURN=0
	RESULT=
	
	if [ $# -ne 3 ] ; then
		Log "createParentWithOwnershipAndPermissions: wrong number of arguments."
		return 2
	fi
	
	local IN_PATH="$1"
	local IN_OWNER_GROUP="$2"
	local IN_PERMISSIONS="$3"
	
	if [ -z "${IN_PATH}" ] ; then
		Log "createParentWithOwnershipAndPermissions: path must be specified."
		return 2
	fi
	
	if [ -z "${IN_OWNER_GROUP}" ] ; then
		Log "createParentWithOwnershipAndPermissions: owner & group not specified."
		return 2
	fi
	
	if [ -z "${IN_PERMISSIONS}" ] ; then
		Log "createParentWithOwnershipAndPermissions: permissions not specified."
		return 2
	fi
	
	local PARENT_PATH=`${DIRNAME} "${IN_PATH}"`
	RETURN=$?
	if [ ${RETURN} -eq 0 ] ; then
		if [ ! -d "${PARENT_PATH}" ] ; then
			createParentWithOwnershipAndPermissions "${PARENT_PATH}" "${IN_OWNER_GROUP}" "${IN_PERMISSIONS}"
			RETURN=$?
			if [ ${RETURN} -eq 0 ] ; then
				${MKDIR} "${PARENT_PATH}"
				RETURN=$?
				if [ ${RETURN} -eq 0 ] ; then
					${CHOWN} "${IN_OWNER_GROUP}" "${PARENT_PATH}"
					RETURN=$?
					if [ ${RETURN} -eq 0 ] ; then
						${CHMOD} "${IN_PERMISSIONS}" "${PARENT_PATH}"
						if [ ${RETURN} -ne 0 ] ; then
							Log "createParentWithOwnershipAndPermissions: unable to set permissions for ${PARENT_PATH}."
						fi
					else
						Log "createParentWithOwnershipAndPermissions: unable to set owner:group to ${IN_OWNER_GROUP} for ${PARENT_PATH}."
					fi
				else
					Log "createParentWithOwnershipAndPermissions: unable to create directory at ${PARENT_PATH}."
				fi
			fi
		fi
	else
		Log "createParentWithOwnershipAndPermissions: error getting parent path for ${IN_PATH}."
	fi
	
	return ${RETURN}
}



####
# createFolderWithOwnershipAndPermissions inPath inOwnerGroup inPermissions
# Creates the folder and any intermediates with the specified ownership and permissions.
# Ownership and permissions of existing parent folders remains unchanged.  If
# the folder already exists, this does nothing.
#   inPath          path to file/folder to create
#   inOwnerGroup    owner:group to use for created folder and missing parents
#   inPermissions   permissions to apply to newly created folder and missing parents
#   RESULT          unset
#   return          0 on success, non-zero otherwise
function createFolderWithOwnershipAndPermissions
{
	local RETURN=0
	RESULT=
	
	if [ $# -ne 3 ] ; then
		Log "createFolderWithOwnershipAndPermissions: wrong number of arguments."
		return 2
	fi
	
	local IN_PATH="$1"
	local IN_OWNER_GROUP="$2"
	local IN_PERMISSIONS="$3"
	
	if [ -z "${IN_PATH}" ] ; then
		Log "createFolderWithOwnershipAndPermissions: path must be specified."
		return 2
	fi
	
	if [ -z "${IN_OWNER_GROUP}" ] ; then
		Log "createFolderWithOwnershipAndPermissions: owner & group not specified."
		return 2
	fi
	
	if [ -z "${IN_PERMISSIONS}" ] ; then
		Log "createFolderWithOwnershipAndPermissions: permissions not specified."
		return 2
	fi
	
	if [ ! -d "${IN_PATH}" ] ; then
		createParentWithOwnershipAndPermissions "${IN_PATH}" "${IN_OWNER_GROUP}" "${IN_PERMISSIONS}"
		RETURN=$?
		if [ ${RETURN} -eq 0 ] ; then
			${MKDIR} "${IN_PATH}"
			RETURN=$?
			if [ ${RETURN} -eq 0 ] ; then
				${CHOWN} "${IN_OWNER_GROUP}" "${IN_PATH}"
				RETURN=$?
				if [ ${RETURN} -eq 0 ] ; then
					${CHMOD} "${IN_PERMISSIONS}" "${IN_PATH}"
					if [ ${RETURN} -ne 0 ] ; then
						Log "createFolderWithOwnershipAndPermissions: unable to set permissions for ${IN_PATH}."
					fi
				else
					Log "createFolderWithOwnershipAndPermissions: unable to set owner:group to ${IN_OWNER_GROUP} for ${IN_PATH}."
				fi
			else
				Log "createFolderWithOwnershipAndPermissions: unable to create folder at ${IN_PATH}."
			fi
		else
			Log "createFolderWithOwnershipAndPermissions: unable to create parent folders for ${IN_PATH}."
		fi
	fi
	
	return ${RETURN}
}



####
# exitOnError inFunction [parameters...]
# Runs the specified function with its parameters.  If the result of invoking
# function is non-zero, exit is called with that result.
#   inFunction  function to execute and check results of
#   ...         parameters to pass to inFunction
#   RESULT      unset
#   return      0 on success, non-zero otherwise (exit if inFunction result is non-zero)
function exitOnError
{
	local RETURN=0
	RESULT=
	
	if [ $# -lt 1 ] ; then
		Log "exitOnError: at least a function must be specified."
		return 2
	fi
	
	local IN_FUNCTION="$1"
	shift
	
	"${IN_FUNCTION}" "$@"
	local RETURN=$?
	if [ ${RETURN} -ne 0 ] ; then
		Log "exitOnError: exiting due to failure in ${IN_FUNCTION}."
		exit ${RETURN}
	fi
	
	return ${RETURN}
}



# Register or unregister an installed product for AUM updates.
#
# Parameters:
# 
#    1: full path to "AMI" plist to use to construct AUM app ID
#    2: flag indicating whether to register (1) or unregister (0)

function RegisterForUpdates
{
	declare -r AppInstallFolder="$1"
	declare -r AMIPList="$2"
	declare -r DoRegister="$3"
	
    CheckArgumentCount "RegisterForUpdates" $# 3
	
	declare -r INVOKE_ADOBE_UPDATE="${gUpdaterResources}/invokeAdobeUpdate"
	
	declare -r UPDATE_ESD_APPID=`${PLIST_RD} "${AMIPList}" AMIAUMApplicationIdentifier`
	declare -r UPDATE_APP_DISPLAY_NAME=`${PLIST_RD} "${AMIPList}" AMIProductName`
	declare -r UPDATE_APP_PATH="${AppInstallFolder}"
	declare -r UPDATE_LANGUAGE=`${PLIST_RD} "${AMIPList}" AMIDefaultLanguage`
	declare -r UPDATE_PREAPP=""
	declare -r UPDATE_PREAPP_PARMS=""
	declare -r UPDATE_POSTAPP=""
	declare -r UPDATE_POSTAPP_PARMS=""
	declare AUMReleaseType=`${PLIST_RD} "${AMIPList}" AMIAUMReleaseType`
	
	if [ -z "${AUMReleaseType}" ] ; then
		# work-around for VC inconsistency:
		AUMReleaseType="CS"
	fi
	
	declare AUMAppID="${UPDATE_ESD_APPID}-${UPDATE_LANGUAGE}-${AUMReleaseType}"
	Log "Constructed AUM app ID is: ${AUMAppID}"
	
	if [ -x "${INVOKE_ADOBE_UPDATE}" ] ; then
		
		# Check to see if the directory Updater creates exists in the local domain
		if [ -d "/Library/Application Support/Adobe/Updater" ] ; then
			Log "Did not create /Library/Application Support/Adobe/Updater because it already exists."
		else
			# If it does not exist, create directory with root:admin ownership:
			${MKDIR} "/Library/Application Support/Adobe/Updater"
		fi

		# Check to see if the directory Updater creates exists in the user domain
		if [ -d "~/Library/Application Support/Adobe/Updater" ] ; then
			Log "Did not create ~/Library/Application Support/Adobe/Updater because it already exists."
		else
			# If it does not exist, create user domain updater directory with correct ownership:
			${SUDO} -u "${USERNAME}" ${MKDIR} "~/Library/Application Support/Adobe/Updater"
		fi
		
		declare AUMAction
		if [ "${DoRegister}" -eq 1 ] ; then
			"${INVOKE_ADOBE_UPDATE}" "register" "${AUMAppID}" "${UPDATE_APP_DISPLAY_NAME}" "${UPDATE_APP_PATH}" "${UPDATE_PREAPP}" "${UPDATE_PREAPP_PARMS}" "${UPDATE_POSTAPP}" "${UPDATE_POSTAPP_PARMS}"
			declare Err=$?
		elif [ "${DoRegister}" -eq 0 ] ; then
			"${INVOKE_ADOBE_UPDATE}" "unregister" "${AUMAppID}"
			declare Err=$?
		else
			FailWithError 1 "Error: incorrect option passed to RegisterForUpdates"
		fi
		
		if [ ${Err} -ne 5 ] ; then	# 5 = already registered
			Log "error registering with AUM: ${Err}"
# *** too stringent...
#			FailWithError "${Err}" "invokeAdobeUpdate failed for ${UPDATE_APP_DISPLAY_NAME}"
		fi
	else
		Log "error: invokeAdobeUpdate not found or not executable"
	fi
}



# Verifies that a specified volume has at least the requested free space.
#
# Parameters:
# 
#    1: path of volume on which to check free space (arbitrary paths allowed; host device checked)
#    2: required space, specified in 1K blocks
# 
# Result codes:
# 
#    0: sufficient space available
#    1: insufficient space available

function CheckVolumeSpace
	{
	declare -r CheckVolumePath="$1"
	declare -r Required1KBlocks="$2"
	
    CheckArgumentCount "CheckVolumeSpace" $# 2
	
	declare -i VolCheckResult=0
	
	Log "Checking space on: ${CheckVolumePath}"
	Log "Required blocks: ${Required1KBlocks}"
	
	declare -r -i VolFreeBlocks=`${DF} -k "${CheckVolumePath}" | ${SED} -n '2p' | ${AWK} '{print $4}'`
	Log "VolFreeBlocks: ${VolFreeBlocks}"
	
	if [ "${Required1KBlocks}" -gt "${VolFreeBlocks}" ] ; then
		VolCheckResult=1
	fi
	
	Log "space check result: ${VolCheckResult}"
	
	return "${VolCheckResult}"
	}



